#ifndef __COMMON_H__
#define __COMMON_H__
#include <string.h>
#include <string>
#include <stdio.h>
#include <assert.h>
#include <stdint.h>
#include <math.h>
#include <algorithm>
#include <list>
#include <vector>
#include <complex>
#include <unistd.h>
#include <iostream>
#include <limits.h>
#include "mpi.h"
namespace CTF {

  /**
   * labels corresponding to symmetry of each tensor dimension
   * NS = 0 - nonsymmetric
   * SY = 1 - symmetric
   * AS = 2 - antisymmetric
   * SH = 3 - symmetric hollow
   */
  #if (!defined NS && !defined SY && !defined SH)
  #define NS 0
  #define SY 1
  #define AS 2
  #define SH 3
  #endif


  /**
   * \brief reduction types for tensor data
   *        deprecated types: OP_NORM1=OP_SUMABS, OP_NORM2=call norm2(), OP_NORM_INFTY=OP_MAXABS
   */
  enum OP { OP_SUM, OP_SUMABS, OP_SUMSQ, OP_MAX, OP_MIN, OP_MAXABS, OP_MINABS};

  enum { SUCCESS, ERROR, NEGATIVE };

  template <typename type=char>
  int conv_idx(int          order,
               type const * cidx,
               int **       iidx);

  template <typename type=char>
  int conv_idx(int          order_A,
               type const * cidx_A,
               int **       iidx_A,
               int          order_B,
               type const * cidx_B,
               int **       iidx_B);

  template <typename type=char>
  int conv_idx(int          order_A,
               type const * cidx_A,
               int **       iidx_A,
               int          order_B,
               type const * cidx_B,
               int **       iidx_B,
               int          order_C,
               type const * cidx_C,
               int **       iidx_C);


}


namespace CTF_int {
  //latency time per message
  #define COST_LATENCY (1.e-6)
  //memory bandwidth: time per per byte
  #define COST_MEMBW (1.e-9)
  //network bandwidth: time per byte
  #define COST_NETWBW (5.e-10)
  //flop cost: time per flop
  #define COST_FLOP (2.e-11)
  //flop cost: time per flop
  #define COST_OFFLOADBW (5.e-10)
  //max total virtualization factor for mappings
  #define MAX_NVIRT 256
  //min total virtualization factor for mappings 
  // (would be useful if explicit blockwise threading was enabled, which is not currently)
  #ifndef MIN_NVIRT
  #define MIN_NVIRT 1
  #endif

  void flops_add(int64_t n);

  int64_t get_flops();
/*
  void csgemm(char transa,      char transb,
              int m,            int n,
              int k,            float a,
              float const * A,  int lda,
              float const * B,  int ldb,
              float b,          float * C,
                                int ldc);

  void cdgemm(char transa,      char transb,
              int m,            int n,
              int k,            double a,
              double const * A, int lda,
              double const * B, int ldb,
              double b,         double * C,
                                int ldc);

  void ccgemm(char transa,                    char transb,
              int m,                          int n,
              int k,                          std::complex<float> a,
              const std::complex<float> * A,  int lda,
              const std::complex<float> * B,  int ldb,
              std::complex<float> b,          std::complex<float> * C,
                                              int ldc);


  void czgemm(char transa,                    char transb,
              int m,                          int n,
              int k,                          std::complex<double> a,
              const std::complex<double> * A, int lda,
              const std::complex<double> * B, int ldb,
              std::complex<double> b,         std::complex<double> * C,
                                              int ldc);

  void csaxpy(int n,              float  dA,
              const float  * dX,  int incX,
              float  * dY,        int incY);

  void cdaxpy(int n,              double dA,
              const double * dX,  int incX,
              double * dY,        int incY);

  void ccaxpy(int n,                            std::complex<float> dA,
              const std::complex<float> * dX,   int incX,
              std::complex<float> * dY,         int incY);

  void czaxpy(int n,                            std::complex<double> dA,
              const std::complex<double> * dX,  int incX,
              std::complex<double> * dY,        int incY);


  void csscal(int n, float dA, float * dX, int incX);

  void cdscal(int n, double dA, double * dX, int incX);

  void ccscal(int n, std::complex<float> dA, std::complex<float> * dX, int incX);

  void czscal(int n, std::complex<double> dA, std::complex<double> * dX, int incX);
*/

  class CommData {
    public:
      MPI_Comm cm;
      int np;
      int rank;
      int color;
      int alive;
      int created;

      //CommData();
      ~CommData();

      /** \brief copy constructor sets created to zero */
      CommData(CommData const & other);
      CommData& operator=(CommData const & other);

      /**
       * \brief create active communicator wrapper
       * \param[in] cm MPI_Comm defining this wrapper
       */
      CommData(MPI_Comm cm);

      /**
       * \brief create non-active communicator wrapper
       * \param[in] rank rank within this comm
       * \param[in] color identifier of comm within parent
       * \param[in] np number of processors within this comm
       */
      CommData(int rank, int color, int np);

      /**
       * \brief create active subcomm from parent comm which must be active
       * \param[in] rank processor rank within subcomm
       * \param[in] color identifier of subcomm within this comm
       * \param[in] parent comm to split
      */
      CommData(int rank, int color, CommData parent);

      /**
       * \brief activate this subcommunicator by splitting parent_comm
       * \param[in] parent_comm communicator to split
       */
      void activate(MPI_Comm parent);

      /* \brief deactivate (MPI_Free) this comm */
      void deactivate();
     
      /* \brief provide estimate of broadcast execution time */
      double estimate_bcast_time(int64_t msg_sz);
   
      /* \brief provide estimate of allreduction execution time */
      double estimate_allred_time(int64_t msg_sz);
     
      /* \brief provide estimate of all_to_all execution time */
      double estimate_alltoall_time(int64_t chunk_sz);
     
      /* \brief provide estimate of all_to_all_v execution time */
      double estimate_alltoallv_time(int64_t tot_sz);

      /**
       * \brief performs all-to-all-v with 64-bit integer counts and offset on arbitrary
       *        length types (datum_size), and uses point-to-point when all-to-all-v sparse
       * \param[in] send_buffer data to send
       * \param[in] send_counts number of datums to send to each process
       * \param[in] send_displs displacements of datum sets in sen_buffer
       * \param[in] datum_size size of MPI_datatype to use
       * \param[in,out] recv_buffer data to recv
       * \param[in] recv_counts number of datums to recv to each process
       * \param[in] recv_displs displacements of datum sets in sen_buffer
       */
      void all_to_allv(void *           send_buffer, 
                       int64_t const * send_counts,
                       int64_t const * send_displs,
                       int64_t          datum_size,
                       void *           recv_buffer, 
                       int64_t const * recv_counts,
                       int64_t const * recv_displs);

  };

  #ifndef ENABLE_ASSERT
  #ifdef DEBUG
  #define ENABLE_ASSERT 1
  #else
  #define ENABLE_ASSERT 0
  #endif
  #endif

  void handler();

  #ifndef ASSERT
  #if ENABLE_ASSERT
  #define ASSERT(...)                \
  do { if (!(__VA_ARGS__)) CTF_int::handler(); assert(__VA_ARGS__); } while (0)
  #else
  #define ASSERT(...) do {} while(0 && (__VA_ARGS__))
  #endif
  #endif

  int  alloc_ptr(int64_t len, void ** const ptr);
  int  mst_alloc_ptr(int64_t len, void ** const ptr);
  void * alloc(int64_t len);
  void * mst_alloc(int64_t len);
  int cfree(void * ptr, int const tid);
  int cfree(void * ptr);
}

/*
inline int CTF_alloc_ptr(int64_t len, void ** const ptr){
  int pm = posix_memalign(ptr, 16, len);
  if (pm) return CTF::ERROR;
  else return CTF::SUCCESS;
}

inline int CTF_mst_alloc_ptr(int64_t len, void ** const ptr){
  return CTF_alloc_ptr(len, ptr);
}
inline void * CTF_alloc(int64_t len){
  void * ptr;
  int pm = posix_memalign(&ptr, 16, len);
  if (pm){ ASSERT(0); return NULL; }
  else return ptr;
}
inline void * CTF_mst_alloc(int64_t len){
  return CTF_alloc(len);
}
inline void CTF_int::cfree(void * ptr, int const tid){
  free(ptr);
}
inline void CTF_int::cfree(void * ptr){
  free(ptr);
}
*/

#endif
